/**
*
* @package VorteX vFOSe.ru IpUpdater
* @version $Id: socklib.cpp,v 0.1 2007/09/29 15:49:51 nE0sIghT Exp $
* @copyright (c) 2007 nE0sIghT
* @license GNU General Public Lecense v2
*
*/


#include "socklib.h"

socklib::socklib(bool StartWSA)
{
#ifdef WIN32
	stopWSA = false;
	if(StartWSA)
	{
		stopWSA = false;
		if(StartWSA)
		{
			if(WSAStartup(MAKEWORD(2, 0), &WsaData) != 0)
			{
				Error(SLE_WSASTARTUP);
				return;
			}
			stopWSA = true;
		}
	}
#endif

	cbError = 0;
	cbRecv = 0;
	sock = cSock = 0;
}

socklib::~socklib()
{
	Stop();

#ifdef WIN32
	if(stopWSA)
	{
		if(WSACleanup() != 0)
		{
			Error(SLE_WSACLEANUP);
		}
	}
#endif
}

void socklib::SetUp(SL_SET sl_set)
{
	memcpy(&set, &sl_set, sizeof(SL_SET));
}

void socklib::SetCallbacks(pSLErrorCallback ec, pSLRecvCallback rc, pSLAcceptCallback ac)
{
	if(ec != NULL)
		cbError = ec;
	if(rc != NULL)
		cbRecv = rc;
	if(ac != NULL)
		cbAccept = ac;
}

void socklib::Start()
{
	if((sock = socket(AF_INET, (set.PROTO == SL_UDP ? SOCK_DGRAM : SOCK_STREAM), (set.PROTO == SL_UDP ? IPPROTO_UDP : IPPROTO_TCP))) != INVALID_SOCKET)
	{
		struct sockaddr_in sin = {0};

		if(set.BindPort)
		{
			sin.sin_addr.s_addr = (strlen(set.ClientIp) ? inet_addr(set.ClientIp) : INADDR_ANY);
			sin.sin_port = htons(set.BindPort);
			sin.sin_family = AF_INET;

			if(bind(sock, (struct sockaddr*)&sin, sizeof(sin)) == SOCKET_ERROR)
			{
				Error(SLE_BIND);
			}

			if(set.ServerIP && !set.ServerPort)
			{
				if(listen(sock, SOMAXCONN) == SOCKET_ERROR)
				{
					Error(SLE_LISTEN);
				}
			}
		}

		if(strlen(set.ServerIP) && set.ServerPort)
		{
			sin.sin_addr.s_addr = inet_addr(set.ServerIP);
			sin.sin_port = htons(set.ServerPort);
			sin.sin_family = AF_INET;

			if(connect(sock, (struct sockaddr*)&sin, sizeof(sin)) == SOCKET_ERROR)
			{
				Error(SLE_CONNECT);
			}
		}
	}
	else
		Error(SLE_INVALID_SOCKET);
}

void socklib::Stop()
{
	if(cSock)
	{
		shutdown(cSock, SD_BOTH);
		closesocket(cSock);

		cSock = 0;
	}
	if(sock)
	{
		shutdown(sock, SD_BOTH);
		closesocket(sock);

		sock = 0;
	}
}

void socklib::Send(char *packet, size_t len)
{
	SOCKET rSock = cSock ? cSock : sock;
	if(rSock)
	{
		if(send(rSock, packet, len, 0) == SOCKET_ERROR)
		{
			Error(SLE_SEND);
		}
	}
}

void socklib::Recv()
{
	char buff[8192], ip[16] = {0};
	size_t len = sizeof(buff);

	RecvEx(buff, &len, ip);

	if(cbRecv && len)
		cbRecv(buff, len, ip);
}

char* socklib::RecvEx(char *buff, size_t* len, char *fromBuff)
{
	SOCKET rSock = cSock ? cSock : sock;
	if(rSock)
	{
		fd_set fd_read = {0};
		FD_ZERO(&fd_read);
		timeval tv = {0};
		struct sockaddr_in sin;
		socklen_t sinSize;

		FD_SET(rSock, &fd_read);
		if(set.so_recvtimeo > 1000)
			tv.tv_sec = set.so_recvtimeo / 1000;
		tv.tv_usec = (set.so_recvtimeo - (set.so_recvtimeo / 1000)*1000);
		sinSize = sizeof(sin);
		ZeroMemory(&sin, sizeof(sin));

		if(select(0, &fd_read, NULL, NULL, &tv))
		{
			ZeroMemory(buff, *len);
			if((*len = recvfrom(rSock, buff, *len, 0, (struct sockaddr*)&sin, &sinSize)) == SOCKET_ERROR)
			{
				Error(SLE_RECV);
			}
			if(*len > 0 && fromBuff)
			{
				sprintf(fromBuff, inet_ntoa(sin.sin_addr));
			}
		}
		else
			*len = 0;
	}

	return buff;
}

void socklib::Accept(char *fromBuff)
{
	if(sock)
	{
		cSock = 0;

		fd_set fd_read = {0};
		FD_ZERO(&fd_read);
		timeval tv = {0};
		struct sockaddr_in sin;
		socklen_t sinSize;

		FD_SET(sock, &fd_read);
		if(set.so_recvtimeo > 1000)
			tv.tv_sec = set.so_recvtimeo / 1000;
		tv.tv_usec = (set.so_recvtimeo - (set.so_recvtimeo / 1000)*1000);
		sinSize = sizeof(sin);
		ZeroMemory(&sin, sizeof(sin));

		if(select(0, &fd_read, NULL, NULL, &tv))
		{
			if((cSock = accept(sock, (struct sockaddr*)&sin, &sinSize)) == SOCKET_ERROR)
			{
				Error(SLE_ACCEPT);
			}
			else
			{
				sprintf(fromBuff, inet_ntoa(sin.sin_addr));
				cbAccept(fromBuff);
			}
		}
	}
}

void socklib::RealError(int errcode, char *file, int line)
{
	Stop();
	if(cbError)
	{
		char StrErr[2048];
		char buff[1024] = {0};

#ifdef WIN32
		FormatMessage(
			FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			GetLastError(),
			LANG_SYSTEM_DEFAULT,
			buff,
			sizeof(buff),
			NULL
		);
#else
		sprintf(buff, "%s", strerror(errcode));
#endif

		sprintf(StrErr, "%s", buff);
		cbError(errcode, StrErr, file, line);
	}
}
